@@ -12,7 +12,7 @@ August 21, 2019
 import time
 import io
 
-from typing import Dict, Optional, Union, List
+from typing import Dict, Optional, Union, List, Tuple
 
 import requests
 
@@ -23,17 +23,23 @@ from .query_structures import *
 from .authentication import get_api_token
 from .exceptions import APIQueryFailedException
 
-__all__ = ['SDAPIQuery', 'SDJobRunner']
+__all__ = ['SDAPIQuery', 'SDJobRunner', 'api_configuration']
 
 T_DF = pandas.DataFrame
 
 
+api_configuration = {
+    'chunksize': 500,
+    'timeout': 60.0,
+}
+
+
 class SDAPIQuery:
     @classmethod
     def submit_query(cls,
                      query_params: Optional[APIQueryParams] = None,
                      queries_params: Optional[Dict[str, APIQueryParams]] = None,
-                     timeout: Optional[float] = 60.0,
+                     timeout: Optional[float] = None,
                      headers: Optional[Dict[str, str]] = None) -> Union[T_DF, Dict[str, T_DF]]:
         """
         Determines the proper method to use and passes values along for request submission
@@ -59,6 +65,9 @@ class SDAPIQuery:
         assert not (query_params is None and queries_params is None)
         assert not (query_params is not None and queries_params is not None)
 
+        if timeout is None:
+            timeout = api_configuration['timeout']
+
         if query_params is not None:
             return cls.query_api_df(query_params=query_params, headers=headers, timeout=timeout)
         elif queries_params is not None:
@@ -66,7 +75,7 @@ class SDAPIQuery:
 
     @staticmethod
     def query_api_json(query_params: APIQueryParams,
-                       timeout: Optional[float] = 60.0,
+                       timeout: Optional[float] = None,
                        headers: Optional[Dict[str, str]] = None) -> dict:
         """
         Submits the query params and returns the resulting data
@@ -83,6 +92,9 @@ class SDAPIQuery:
         -------
         A dict containing the query result
         """
+        if timeout is None:
+            timeout = api_configuration['timeout']
+
         json_data = _submit_post_request(
             json_dict=dict(token=get_api_token(), query=query_params.to_api_struct()),
             headers=headers,
@@ -92,7 +104,7 @@ class SDAPIQuery:
 
     @staticmethod
     def query_api_df(query_params: APIQueryParams,
-                     timeout: Optional[float] = 60.0,
+                     timeout: Optional[float] = None,
                      headers: Optional[Dict[str, str]] = None) -> pandas.DataFrame:
         """
         Submits the query params and returns the resulting data
@@ -110,6 +122,9 @@ class SDAPIQuery:
         -------
         A pandas dataframe containing the query result
         """
+        if timeout is None:
+            timeout = api_configuration['timeout']
+
         json_data = _submit_post_request(
             json_dict=dict(token=get_api_token(), query=query_params.to_api_struct()),
             headers=headers,
@@ -122,8 +137,8 @@ class SDAPIQuery:
 
     @staticmethod
     def query_api_multiple(queries: Dict[str, APIQueryParams],
-                           timeout: Optional[float] = 60.0,
-                           chunksize: int = 500,
+                           timeout: Optional[float] = None,
+                           chunksize: int = None,
                            time_between_chunks: Optional[float] = None,
                            headers: Optional[Dict[str, str]] = None) -> Dict[str, pandas.DataFrame]:
         """
@@ -146,6 +161,12 @@ class SDAPIQuery:
         -------
         A dict with pandas dataframes as the values for each of the query params in the input dict
         """
+        if chunksize is None:
+            chunksize = api_configuration['chunksize']
+
+        if timeout is None:
+            timeout = api_configuration['timeout']
+
         assert isinstance(chunksize, int) and chunksize > 0, \
             f'Chunksize must be a positive integer, is {chunksize}'
 
@@ -198,13 +219,38 @@ class SDJobRunner:
                                   geolevel: Optional[str] = None,
                                   response_format: str = 'csv',
                                   portfolio_id: Optional[str] = None,
+                                  market_id: Optional[str] = None,
+                                  sites: Optional[List[Tuple[float, float, str]]] = None,
                                   buffers: Optional[List[str]] = None,
                                   geoid_list: Optional[List[int]] = None) -> pandas.DataFrame:
+        """
+        This method handles all steps of the job pipeline to produce a dataframe, or error if the
+        job fails. From a client's request, this method will start a job, and repeatedly check its
+        status until it finishes, fails, or the maximum wait time expires. If the job completes,
+        this method will download the resulting dataframe and return it to the calling client.
+
+        Parameters
+        ----------
+        model_id
+        geolevel
+        response_format
+        portfolio_id
+        market_id
+        sites
+        buffers
+        geoid_list
+
+        Returns
+        -------
+        pandas.DataFrame
+        """
         self.create_job(
             model_id=model_id,
             geolevel=geolevel,
             response_format=response_format,
             portfolio_id=portfolio_id,
+            market_id=market_id,
+            sites=sites,
             buffers=buffers,
             geoid_list=geoid_list)
 
@@ -226,12 +272,35 @@ class SDJobRunner:
                    model_id: str,
                    geolevel: Optional[str] = None,
                    response_format: str = 'csv',
+                   sites: Optional[List[Tuple[float, float, str]]] = None,
                    portfolio_id: Optional[str] = None,
+                   market_id: Optional[str] = None,
                    geoid_list: Optional[List[int]] = None,
                    buffers: Optional[List[str]] = None) -> None:
+        """
+        This method handles the call to create a job through the API, and will log the result of its
+        request.
+
+        Parameters
+        ----------
+        model_id
+        geolevel
+        response_format
+        sites
+        portfolio_id
+        market_id
+        geoid_list
+        buffers
+
+        Returns
+        -------
+        None
+        """
         assert isinstance(model_id, str), f'model_id must be str (was {model_id})'
         assert portfolio_id is None or isinstance(portfolio_id, str), \
             f'portfolio_id must be str (was {portfolio_id})'
+        assert market_id is None or isinstance(market_id, str), \
+            f'market_id must be str (was {market_id})'
         assert geolevel is None or isinstance(geolevel, str), \
             f'geolevel must be str (was {geolevel})'
         assert response_format in {'csv', 'json'}
@@ -240,15 +309,19 @@ class SDJobRunner:
         assert buffers is None or isinstance(buffers, list)
         assert buffers is None or all(buffer in cc.BUFFERS_TUPLE for buffer in buffers), \
             f'Invalid buffers: {buffers} (must be from {cc.BUFFERS_TUPLE})'
+        assert sites is None or isinstance(sites, list)
+        assert sites is None or all(isinstance(site, tuple) for site in sites)
 
-        # Must either be a geolevel, geoid_list combo or a portfolio_id
-        if not geolevel and not portfolio_id:
-            raise ValueError('Job requires either "geolevel" or "portfolio_id"')
+        # Must either be a geolevel, geoid_list combo, or a portfolio_id, or market_id
+        if not geolevel and not any([portfolio_id, market_id]):
+            raise ValueError('Job requires one of "geolevel", "portfolio_id", or "market_id"')
 
         if geolevel:
-            assert geolevel in {'US', 'METRO', 'GEOID2', 'GEOID5', 'ZIP', 'GEOID11'}, \
+            assert geolevel in {
+                'US', 'METRO', 'GEOID2', 'GEOID5', 'ZIP', 'GEOID11', 'site-addresses'}, \
                 '"geolevel" must be one of "US", "METRO", "GEOID2", "GEOID5", "ZIP", "GEOID11"'
             assert portfolio_id is None, 'Cannot have both "geolevel" and "portfolio_id"'
+            assert market_id is None, 'Cannot have both "geolevel" and "market_id"'
 
         if geoid_list is None:
             geoid_list = []
@@ -256,6 +329,9 @@ class SDJobRunner:
         if buffers is None:
             buffers = []
 
+        if sites is None:
+            sites = []
+
         if self._logging:
             print('Sending create job request to API service')
 
@@ -267,10 +343,12 @@ class SDJobRunner:
             json=dict(
                 model_id=model_id,
                 portfolio_id=portfolio_id,
+                market_id=market_id,
                 geolevel=geolevel,
                 response_format=response_format,
                 geoid_list=geoid_list,
                 buffers=buffers,
+                sites=sites,
             )
         )
 
@@ -285,6 +363,14 @@ class SDJobRunner:
             raise APIQueryFailedException(res['message'])
 
     def _check_job_status(self) -> str:
+        """
+        This method will request the status of the job related to an instance of this class and
+        return the status string.
+
+        Returns
+        -------
+        str
+        """
         self._assert_job_created()
 
         r = requests.post(
@@ -306,6 +392,14 @@ class SDJobRunner:
             return r['message']
 
     def download_job_to_dataframe(self) -> pandas.DataFrame:
+        """
+        This method will handle the API request to download the job related to an instance of this
+        class, and return the dataframe
+
+        Returns
+        -------
+        pandas.DataFrame
+        """
         self._assert_job_created()
 
         r = requests.post(
